/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "string_bt.hpp"
#include <string_view>
#include <cstdint>

namespace wren {
	namespace detail {
		template <dhandy::bt::string S>
		struct ModuleAndNameStaticStorage {
			static constexpr const auto value = S;
		};
	} //unnamed namespace

	//Non-owning, lightweight class for storing a module nome/class name pair.
	//They are stored as consecutive strings within a single buffer with a
	//null-char in-between and at the end so that each name is null-terminated
	//and can be used with C APIs.
	//Keep the size small, ideally within 2 registers size. Altering this might
	//require changes to the assembly source files for DynafuncMaker.
	//When passed as a parameter by value, objects of ModuleAndName fit into
	//2 registers on 64 bit architectures. This is important because objects
	//of this class get hardcoded in the dynamic functions that DynafuncMaker
	//spits out.
	class ModuleAndName {
	public:
		constexpr ModuleAndName (const char* base, std::uint32_t hash, std::size_t mod_name_len, std::size_t class_name_len);

		constexpr const char* module_name_char() const noexcept { return m_base; }
		constexpr const char* class_name_char() const noexcept { return m_base + 1 + m_mod_name_len; }
		constexpr std::size_t module_name_len() const noexcept { return m_mod_name_len; }
		constexpr std::size_t class_name_len() const noexcept { return m_class_name_len; }
		constexpr std::string_view module_name() const noexcept { return {module_name_char(), m_mod_name_len}; }
		constexpr std::string_view class_name() const noexcept { return {class_name_char(), m_class_name_len}; }
		constexpr const char* buffer_address() const noexcept { return m_base; }

	private:
		const char* m_base;
		std::uint32_t m_hash;
		std::uint16_t m_mod_name_len;
		std::uint16_t m_class_name_len;
	};

	union RawAccessModuleAndName {
		constexpr RawAccessModuleAndName (ModuleAndName mn) :
			module_and_name(mn)
		{ }

		ModuleAndName module_and_name;

		struct {
			void* string;
			void* data;
		} raw;

		static_assert(sizeof(raw) == sizeof(module_and_name));
	};

	constexpr ModuleAndName::ModuleAndName (const char* base, std::uint32_t hash, std::size_t mod_name_len, std::size_t class_name_len) :
		m_base(base),
		m_hash(hash),
		m_mod_name_len(mod_name_len),
		m_class_name_len(class_name_len)
	{}

	template <dhandy::bt::string S1, dhandy::bt::string S2>
	constexpr ModuleAndName make_module_and_name() {
		using dhandy::bt::string;

		constexpr const char* data = detail::ModuleAndNameStaticStorage<S1 + string("\0") + S2>::value.data();
		constexpr std::uint16_t s1_len = static_cast<std::uint16_t>(S1.size());
		constexpr std::uint16_t s2_len = static_cast<std::uint16_t>(S2.size());
		constexpr std::uint32_t hash = 0; //not implemented yet

		return ModuleAndName{data, hash, s1_len, s2_len};
	}

	//short-hand alias for make_module_and_name
	template <dhandy::bt::string S1, dhandy::bt::string S2>
	constexpr ModuleAndName MN = make_module_and_name<S1, S2>();
} //namespace wren
