/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "vm.hpp"
#include "detail/string_bt.hpp"
#if defined(WRENPP_WITH_NAME_GUESSING)
#	include "detail/guess_class_name.hpp"
#endif
#include "detail/wren_types.hpp"
#include "detail/construct_foreign_class.hpp"
#include "detail/setters_getters.hpp"
#include <string>
#include <string_view>
#include <type_traits>
#include <cstddef>
#include <utility>

namespace wren {
	template <typename R, typename... Args>
	R call (VM& vm, const Handle& object, const Handle& method, const Args&... args);

	template <typename R, typename... Args>
	R call (VM& vm, ModuleAndName object, const Handle& method, const Args&... args);

	template <typename R, std::size_t N, typename... Args>
	R call (VM& vm, const Handle& object, const char (&method)[N], const Args&... args);

	template <typename R, std::size_t N, typename... Args>
	R call (VM& vm, ModuleAndName object, const char (&method)[N], const Args&... args);

	template <typename T, typename... Args>
	T* make_wren_object(VM& vm, ModuleAndName mn, Args&&... args);

#if defined(WRENPP_WITH_NAME_GUESSING)
	template <typename T, typename... Args>
	T* make_wren_object(VM& vm, const char* module, Args&&... args);
#endif

	template <auto V>
	foreign_method_t make_method_bindable();

	void interpret (VM& vm, const std::string& module_name, const std::string& script);
	void variable(VM& vm, ModuleAndName mod_and_name, int slot);
	void variable(VM& vm, const Handle& handle, int slot);
	void variable_ensure_slot(VM& vm, ModuleAndName mod_and_name, int slot);
	void variable_ensure_slot(VM& vm, const Handle& handle, int slot);

	namespace detail {
		template <typename T>
		inline void set_single_for_call (VM& vm, int slot, const T& value) {
			set(vm, slot, value);
		}

		template <>
		inline void set_single_for_call (VM& vm, int slot_num, const Handle& handle) {
			vm.set_slot_handle(handle, slot_num);
		}

		template <>
		inline void set_single_for_call (VM& vm, int slot_num, const ModuleAndName& name) {
			variable(vm, name, slot_num);
		}

		//Sets arguments so wren is ready to perform a function call. It doesn't
		//touch at slot 0, so eventual objects for method calls must be set
		//manually
		template <typename... Args, int... Indices>
		inline void set_for_call (std::integer_sequence<int, Indices...>, VM& vm, const Args&... args) {
			vm.ensure_slots(sizeof...(Args));
			(set_single_for_call(vm, Indices + 1, args), ...);
		}

		template <typename T, typename R, auto Method, typename... Args> struct MakeMethodBindableConstNonConst {
		private:
			template <int... Indices>
			static R call_implem (std::integer_sequence<int, Indices...>, VM& vm) {
				static_assert(sizeof...(Indices) == sizeof...(Args), "Mismatching argument count");
				T* obj = foreign<T>(vm, 0);
				return (obj->*Method)(get<std::decay_t<Args>>(vm, Indices + 1)...);
			}

		public:
			static foreign_method_t make() {
				return [](VM& vm, ModuleAndName) {
					if constexpr (std::is_same_v<R, void>) {
						call_implem(std::make_integer_sequence<int, sizeof...(Args)>(), vm);
					}
					else {
						auto ret = call_implem(std::make_integer_sequence<int, sizeof...(Args)>(), vm);
						vm.ensure_slots(1);
						set(vm, 0, ret);
					}
				};
			}
		};

		template <auto V> struct MakeMethodBindable;
		template <typename T, void(T::*Method)(VM&)> struct MakeMethodBindable<Method> {
			static foreign_method_t make() {
				return [](VM& vm, ModuleAndName) {
					T* obj = foreign<T>(vm, 0);
					(obj->*Method)(vm);
				};
			}
		};

		template <typename T, typename R, typename... Args, R(T::*Method)(Args...)> struct MakeMethodBindable<Method> : MakeMethodBindableConstNonConst<T, R, Method, Args...> {};
		template <typename T, typename R, typename... Args, R(T::*Method)(Args...)const> struct MakeMethodBindable<Method> : MakeMethodBindableConstNonConst<T, R, Method, Args...> {};
	} //namespace detail

	inline void interpret (VM& vm, const std::string& module_name, const std::string& script) {
		return vm.interpret(module_name.c_str(), script.c_str());
	}

	template <typename R, typename... Args>
	inline R call (VM& vm, const Handle& object, const Handle& method, const Args&... args) {
		detail::set_for_call<Args...>(std::make_integer_sequence<int, sizeof...(Args)>(), vm, args...);
		vm.set_slot_handle(object, 0);
		vm.call(method);
		if constexpr (not std::is_same<void, R>::value) {
			return get<R>(vm, 0);
		}
	}

	template <typename R, typename... Args>
	inline R call (VM& vm, ModuleAndName object, const Handle& method, const Args&... args) {
		vm.ensure_slots(sizeof...(args) + 1);
		variable(vm, object, 0);
		Handle obj_handle = vm.slot_handle(0);
		return call<R, Args...>(vm, obj_handle, method, args...);
	}

	template <typename R, std::size_t N, typename... Args>
	inline R call (VM& vm, const Handle& object, const char (&method)[N], const Args&... args) {
		using dhandy::bt::string;

		const constexpr char dummy_name_buff[N] = {0};
		const constexpr auto params = string<N, char>(dummy_name_buff) +
			string("(") +
			((static_cast<void>(args), string(",_")) + ... + string("")).template substr<1>() +
			string(")");
		;
		char cat_buff[params.size() + 1];
		std::copy(method, method + N - 1, cat_buff);
		std::copy(params.data() + N - 1, params.data() + params.size() + 1, cat_buff + N - 1);
		return call<R, Args...>(vm, object, vm.make_call_handle(cat_buff), args...);
	}

	template <typename R, std::size_t N, typename... Args>
	inline R call (VM& vm, ModuleAndName object, const char (&method)[N], const Args&... args) {
		vm.ensure_slots(sizeof...(args) + 1);
		variable(vm, object, 0);
		Handle obj_handle = vm.slot_handle(0);
		return call<R, N, Args...>(vm, obj_handle, method, args...);
	}

	template <typename T, typename... Args>
	inline T* make_wren_object(VM& vm, ModuleAndName mn, Args&&... args) {
		variable(vm, mn, 0);
		void* const mem = vm.set_slot_new_foreign(0, 0, sizeof(T));
		T* const obj = new(mem) T{std::forward<Args>(args)...};
		return obj;
	}

#if defined(WRENPP_WITH_NAME_GUESSING)
	template <typename T, typename... Args>
	inline T* make_wren_object(VM& vm, const char* module, Args&&... args) {
		return make_wren_object<T>(
			vm,
			ModuleAndName{module, guess_class_name<T>()},
			std::forward<Args>(args)...
		);
	}
#endif

	template <auto V>
	inline foreign_method_t make_method_bindable() {
		return detail::MakeMethodBindable<V>::make();
	}
} //namespace wren
