/* Copyright 2020-2022, Michele Santullo
 * This file is part of wrenpp.
 *
 * Wrenpp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Wrenpp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with wrenpp.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "wrenpp/callback_manager.hpp"
#include <algorithm>
#include <cassert>

namespace wren::detail {
	FullSignatureOwning::FullSignatureOwning (std::string_view module_name, std::string_view class_name, std::string_view signature) :
		FullSignatureBase(std::hash<std::string_view>{}(module_name), std::hash<std::string_view>{}(class_name), std::hash<std::string_view>{}(signature)),
		StringsInVector<3>(module_name, class_name, signature)
	{
	}

	TempFullSignature::TempFullSignature (const FullSignatureOwning& full_sig) :
		FullSignatureBase(full_sig),
		m_module_name(full_sig.module_name()),
		m_class_name(full_sig.class_name()),
		m_signature(full_sig.signature())
	{ }

	TempFullSignature::TempFullSignature (std::string_view module_name, std::string_view class_name, std::string_view signature) :
		FullSignatureBase(std::hash<std::string_view>{}(module_name), std::hash<std::string_view>{}(class_name), std::hash<std::string_view>{}(signature)),
		m_module_name(module_name),
		m_class_name(class_name),
		m_signature(signature)
	{ }

	bool FullSignatureEqual::operator()(TempFullSignature left, TempFullSignature right) const {
		return left.class_name() == right.class_name() and
			left.module_name() == right.module_name() and
			left.signature() == right.signature();
	}

	std::size_t FullSignatureHash::operator() (TempFullSignature key) const {
		//see example at https://en.cppreference.com/w/cpp/utility/hash
		const std::size_t h1 = key.module_hash();
		const std::size_t h2 = key.class_hash();
		const std::size_t h3 = key.signature_hash();
		return h1 ^ ((h2 ^ (h3 << 1)) << 1);
	}
} //namespace wren::detail

namespace wren {
	CallbackManager::CallbackManager() = default;
	CallbackManager::~CallbackManager() noexcept = default;

	CallbackManager& CallbackManager::add_callback (bool is_static, std::string_view module_name, std::string_view class_name, std::string_view signature, foreign_method_t cb) {
		using detail::FullSignatureOwning;
		using detail::TempFullSignature;

		MethodMap& map = *method_map(is_static);
		auto it_found = map.find(TempFullSignature{module_name, class_name, signature});
		if (map.cend() == it_found)
			map.insert(it_found, std::make_pair(FullSignatureOwning{module_name, class_name, signature}, cb));
		else
			it_found->second = cb;

		return *this;
	}

	foreign_method_t CallbackManager::callback (bool is_static, std::string_view module_name, std::string_view class_name, std::string_view signature) const {
		using detail::TempFullSignature;

		const MethodMap* const map = method_map(is_static);
		const auto it_found = map->find(TempFullSignature{module_name, class_name, signature});
		if (map->cend() != it_found)
			return it_found->second;
		else
			return nullptr;
	}

	auto CallbackManager::method_map(bool is_static) noexcept -> MethodMap* {
		return (is_static ? &m_static_methods : &m_normal_methods);
	}

	auto CallbackManager::method_map(bool is_static) const noexcept -> const MethodMap* {
		return const_cast<CallbackManager*>(this)->method_map(is_static);
	}
} //namespace wren
